/* 
   qtfile_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

/*
	Mercilessly ripped from nuvrec (roman@mars.tuwien.ac.at)
	by Justin Schoeman (c) 2001
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "nuppelvideo.h"
#include "minilzo.h"
#include "../RTjpeg/RTjpeg.h"
#include "../NVrec.h"

#define HEAP_ALLOC(var,size) \
	long __LZO_MMODEL var [ ((size) + (sizeof(long) - 1)) / sizeof(long) ]
#define IN_LEN (1024*1024L)
#define OUT_LEN (IN_LEN + IN_LEN / 64 + 16 + 3)

struct nuvfile_core_t {
	int fd;

	int ablocks;
	int vblocks;
	
	RTjpeg_t * rtj;
	struct rtfileheader fileH;
	struct rtframeheader frameH;
	
	unsigned char * cbuf;
	HEAP_ALLOC(wrkmem,LZO1X_1_MEM_COMPRESS);
	lzo_byte out [ OUT_LEN ];
};

struct file_core_t nuvfile_core;


void nuvfile_core_close(struct file_core_t * fct)
{
	struct nuvfile_core_t * nuvc = (struct nuvfile_core_t *)fct->priv;
	if(!nuvc)
		return;
	RTjpeg_close(nuvc->rtj);

	nuvc->fileH.videoblocks = nuvc->vblocks;
	nuvc->fileH.audioblocks = nuvc->ablocks;
//	lseek(nuvc->fd, 0, SEEK_SET);
//	write(nuvc->fd, &nuvc->fileH, sizeof(struct rtfileheader));
	close(nuvc->fd);
	
	free(nuvc);
	free(fct);
}

void nuvfile_core_nullvideo(struct file_core_t * fct)
{
	struct nuvfile_core_t * nuvc = (struct nuvfile_core_t *)fct->priv;
	int i;
	
	if(!nuvc)
		return;
//	fprintf(stderr, "file core null video\n");
	i = RTjpeg_nullcompress(nuvc->rtj, nuvc->cbuf);
	nuvc->frameH.frametype = 'V';
	nuvc->frameH.comptype  = '1';
	nuvc->frameH.packetlength = RTJPEG_COMPAT_LENGTH(nuvc->cbuf);
	write(nuvc->fd, &nuvc->frameH, sizeof(struct rtframeheader));
	write(nuvc->fd, RTJPEG_COMPAT_STREAM(nuvc->cbuf), RTJPEG_COMPAT_LENGTH(nuvc->cbuf));
	
	nuvc->vblocks++;
}

void nuvfile_core_fastvideo(struct file_core_t * fct, unsigned char **buf)
{
	struct nuvfile_core_t * nuvc = (struct nuvfile_core_t *)fct->priv;
	int i;
	
	if(!nuvc)
		return;
//	fprintf(stderr, "file core fast video\n");
	i = RTjpeg_compress(nuvc->rtj, nuvc->cbuf, buf); //planes);
	nuvc->frameH.frametype = 'V';
	nuvc->frameH.comptype  = '1';
	nuvc->frameH.packetlength = RTJPEG_COMPAT_LENGTH(nuvc->cbuf);
	write(nuvc->fd, &nuvc->frameH, sizeof(struct rtframeheader));
	write(nuvc->fd, RTJPEG_COMPAT_STREAM(nuvc->cbuf), RTJPEG_COMPAT_LENGTH(nuvc->cbuf));

	nuvc->vblocks++;
}

void nuvfile_core_video(struct file_core_t * fct, unsigned char **buf)
{
	struct nuvfile_core_t * nuvc = (struct nuvfile_core_t *)fct->priv;
	int i, j, k;
	
	if(!nuvc)
		return;
//	fprintf(stderr, "file core video\n");
	i = RTjpeg_compress(nuvc->rtj, nuvc->cbuf, buf); //planes);
	j = lzo1x_1_compress(RTJPEG_COMPAT_STREAM(nuvc->cbuf), RTJPEG_COMPAT_LENGTH(nuvc->cbuf), nuvc->out, &k, nuvc->wrkmem);
	nuvc->frameH.frametype = 'V';
	nuvc->frameH.comptype  = '2';
	nuvc->frameH.packetlength = k;
	write(nuvc->fd, &nuvc->frameH, sizeof(struct rtframeheader));
	write(nuvc->fd, nuvc->out, k);

	nuvc->vblocks++;
}

void nuvfile_core_audio(struct file_core_t * fct, unsigned char *buf, int n)
{
	struct nuvfile_core_t * nuvc = (struct nuvfile_core_t *)fct->priv;
	
	if(!nuvc)
		return;
//	fprintf(stderr, "file core audio\n");
	nuvc->frameH.frametype = 'A';
	nuvc->frameH.comptype  = '0';
	nuvc->frameH.packetlength = n;
	write(nuvc->fd, &nuvc->frameH, sizeof(struct rtframeheader));
	write(nuvc->fd, buf, n);

	nuvc->ablocks++;
}

void nuvfile_core_sync(struct file_core_t * fct)
{
	struct nuvfile_core_t * nuvc = (struct nuvfile_core_t *)fct->priv;
	fsync(nuvc->fd);
}

int nuvfile_core_format(char **args)
{
/* Parse the command line to find the format that will be needed */
	return NV_YUV420;
}

void nuvfile_core_usage()
{
	fprintf(stderr, "\tnuv file core options:\n");
	fprintf(stderr, "\t-k <key frame rate>  [25]\n");
	fprintf(stderr, "\t-q <quality factor>  [255]\n");
	fprintf(stderr, "\t-l <luma factor>     [2]\n");
	fprintf(stderr, "\t-c <chroma factor>   [1]\n");
	fprintf(stderr, "\t-o <output filename> [NONE]\n");
}

struct file_core_t * nuvfile_core_init(char ** args, ntime_t vrate)
{
	struct file_core_t * fct;
	struct nuvfile_core_t * nuvc;
	double f;
	int i;
 	int arate = -1;
 	int abits = -1;
 	int achans = -1;
 	int vwidth = -1;
 	int vheight = -1;
 	char *fname = NULL;
 	int keyrate = 25;
 	int qfactor = 255;
 	int lfactor = 2;
 	int cfactor = -1;
 	unsigned int tbls[128];
 	NV_ARGS;
 	
 	fct =  malloc(sizeof(struct file_core_t));
 	if(!fct) {
 		perror("nuvfile core init - could no malloc new cap core");
 		return NULL;
 	}
 	nuvc = malloc(sizeof(struct nuvfile_core_t));
 	if(!nuvc) {
 		perror("nuvfile core init - could no malloc new nuv core");
 		free(fct);
 		return NULL;
 	}
 	memcpy(fct, &nuvfile_core, sizeof(struct file_core_t));
 	fct->priv = (void *)nuvc;
 	memset(nuvc, 0, sizeof(struct nuvfile_core_t));
 	
	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-r") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			arate = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-b") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			abits = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-s") == 0) {
			achans = 2;
		} else if(strcmp(NV_THIS_ARG, "-w") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			vwidth = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-h") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			vheight = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-k") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			keyrate = atoi(NV_THIS_ARG);
			if(keyrate < 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-q") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			qfactor = atoi(NV_THIS_ARG);
			if((qfactor < 16) || (qfactor > 255))
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-l") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			lfactor = atoi(NV_THIS_ARG);
			if(lfactor < 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-c") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			cfactor = atoi(NV_THIS_ARG);
			if(cfactor < 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-o") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			fname = NV_THIS_ARG;
		}
	}

	if(cfactor < 0)
		cfactor = lfactor / 2;
	if(!fname)
		goto usage;
	fprintf(stderr, "file core init - file: %s (%dx%d)\n", fname, vwidth, vheight);
	
	nuvc->fd = open(fname, O_RDWR|O_CREAT|O_TRUNC);
	if(!nuvc->fd) {
		perror("file core init - could not open file");
		goto error0;
	}
	f = 1e9 / (double)vrate;
	fprintf(stderr, "file core init - frame rate is %f\n", f);

	nuvc->rtj = RTjpeg_init();
	if(!nuvc->rtj)
		goto error1;
	RTjpeg_set_size(nuvc->rtj, &vwidth, &vheight);
	RTjpeg_set_quality(nuvc->rtj, &qfactor);
	i = RTJ_YUV420;
	RTjpeg_set_format(nuvc->rtj, &i);
	RTjpeg_set_intra(nuvc->rtj, &keyrate, &lfactor, &cfactor);
	
	nuvc->cbuf = malloc(vwidth * vheight * 3);
	if(!nuvc->cbuf) {
		perror("file core init - could not allocate compression buffer");
		goto error2;
	}
	
	if(lzo_init() != LZO_E_OK) {
		fprintf(stderr, "file core init - could not init lzo - falling back to standard RTjpeg\n");
		fct->file_core_video = nuvfile_core_fastvideo;
	}
	
	snprintf(nuvc->fileH.finfo, 12, "NuppelVideo");
	snprintf(nuvc->fileH.version, 5, "0.04");
	nuvc->fileH.width = vwidth;
	nuvc->fileH.height = vheight;
	nuvc->fileH.desiredwidth = 0;
	nuvc->fileH.desiredheight = 0;
	nuvc->fileH.pimode = 'P';
	nuvc->fileH.aspect = 1.0;
	nuvc->fileH.fps = f;
	nuvc->fileH.videoblocks = -1;
	nuvc->fileH.audioblocks = -1;
	nuvc->fileH.textsblocks = 0;
	write(nuvc->fd, &nuvc->fileH, sizeof(struct rtfileheader));
	
	RTjpeg_get_tables(nuvc->rtj, tbls);
	
	nuvc->frameH.frametype = 'D';
	nuvc->frameH.comptype  = 'R';
	nuvc->frameH.packetlength = sizeof(tbls);
	write(nuvc->fd, &nuvc->frameH, sizeof(struct rtframeheader));
	write(nuvc->fd, tbls, sizeof(tbls));
	
	return (void *)fct;

error2:	
	RTjpeg_close(nuvc->rtj);
error1:
	close(nuvc->fd);
error0:
	free(fct);
	free(nuvc);
	return NULL;

usage:
	fprintf(stderr, "nuvfile core usage:\n");
	fprintf(stderr, "Error parsing options.  Please run without any command line options for\n");
	fprintf(stderr, "detailed usage information.\n");
	free(fct);
	free(nuvc);
	return NULL;
}

struct file_core_t nuvfile_core = {
	file_core_init: nuvfile_core_init,
	file_core_close: nuvfile_core_close,
	file_core_nullvideo: nuvfile_core_nullvideo,
	file_core_fastvideo: nuvfile_core_fastvideo,
	file_core_video: nuvfile_core_video,
	file_core_audio: nuvfile_core_audio,
	file_core_sync: nuvfile_core_sync,
	file_core_format: nuvfile_core_format,
	file_core_usage: nuvfile_core_usage,
	priv: NULL
};
