#include <stdio.h>
#include <string.h>
#include <quicktime/quicktime.h>
#include <quicktime/colormodels.h>
#include <quicktime/funcprotos.h>
#include "../RTjpeg/RTjpeg.h"

typedef struct
{
	int quality;      /* 0 - 100 quality */
	char *output_buffer;    /* Buffer for RTJPEG data */
	char *input_buffer;    /* Buffer for RTJPEG data */
	unsigned char *cplanes[3];  /* Buffer for YUV output */
	unsigned char *dplanes[3];  /* Buffer for YUV output */
	unsigned char **row_pointers;
	RTjpeg_t * compress_struct;
	RTjpeg_t * decompress_struct;
} quicktime_rtjpeg_codec_t;

int quicktime_delete_codec_rtjpeg(quicktime_video_map_t *vtrack)
{
	quicktime_rtjpeg_codec_t *codec = ((quicktime_codec_t*)vtrack->codec)->priv;

//	fprintf(stderr, "\nDELETE\n");

	if(codec->output_buffer)
		free(codec->output_buffer);
	codec->output_buffer = 0;
	if(codec->input_buffer)
		free(codec->input_buffer);
	codec->input_buffer = 0;
	if(codec->cplanes[0])
		free(codec->cplanes[0]);
	codec->cplanes[0] = 0;
	if(codec->dplanes[0])
		free(codec->dplanes[0]);
	codec->dplanes[0] = 0;
	RTjpeg_close(codec->compress_struct);
	RTjpeg_close(codec->decompress_struct);
	free(codec);
	return 0;
}


static int quicktime_writes_colormodel_rtjpeg(quicktime_t *file, 
		int colormodel, 
		int track)
{
	return (colormodel == BC_YUV420P);
}

int qt_extern_encode(quicktime_t *file, int track, unsigned char *input, unsigned char *output)
{
	long bytes = 0;
	quicktime_video_map_t *vtrack = &(file->vtracks[track]);
	quicktime_trak_t *trak = vtrack->track;
	quicktime_rtjpeg_codec_t *codec;
	int w = trak->tkhd.track_width;
	int h = trak->tkhd.track_height;

	codec = ((quicktime_codec_t*)file->vtracks[track].codec)->priv;
	
	if(file->color_model == BC_YUV420P) {
		codec->row_pointers[0] = input;
		codec->row_pointers[2] = codec->row_pointers[0] + w * h;
		codec->row_pointers[1] = codec->row_pointers[2] + (w * h)/4;
		bytes = RTjpeg_compress(codec->compress_struct, output, codec->row_pointers);
	} else {
		fprintf(stderr, "Invalid compress color format!");
		return -1;
	}
	return bytes;
}

static int quicktime_reads_colormodel_rtjpeg(quicktime_t *file, 
		int colormodel, 
		int track)
{
//	fprintf(stderr, "READ_CMODEL, %d\n", colormodel);
	return ((colormodel == BC_YUV420P) ||
		(colormodel == BC_RGB8) ||
		(colormodel == BC_RGB565) ||
		(colormodel == BC_RGB888) ||
		(colormodel == BC_RGBA8888) ||
		(colormodel == BC_BGR888) ||
		(colormodel == BC_BGR8888) );
}

int qt_extern_decode(quicktime_t *file, int track, unsigned long inputsize, unsigned char *input, unsigned char *output)
{
	int result = 0;
	quicktime_video_map_t *vtrack = &(file->vtracks[track]);
	quicktime_rtjpeg_codec_t *codec;
	int i, w, h;
	w = vtrack->track->tkhd.track_width;
	h = vtrack->track->tkhd.track_height;

	codec = ((quicktime_codec_t*)file->vtracks[track].codec)->priv;
	
//	fprintf(stderr, "decode %d\n", file->color_model);

	
	if (file->color_model == BC_YUV420P)
	{
		codec->row_pointers[0] = output;
		codec->row_pointers[2] = codec->row_pointers[0] + w * h;
		codec->row_pointers[1] = codec->row_pointers[2] + (w * h)/4;
		RTjpeg_decompress(codec->decompress_struct, input, codec->row_pointers);
	} else {
		RTjpeg_decompress(codec->decompress_struct, input, codec->dplanes);
		switch (file->color_model) {
			case BC_RGB8:
				for(i = 0; i < h; i++)
					codec->row_pointers[i] = output + i * w;
				RTjpeg_yuv420rgb8(codec->decompress_struct, codec->dplanes, codec->row_pointers);
				break;
			case BC_RGB565:
				for(i = 0; i < h; i++)
					codec->row_pointers[i] = output + i * w * 2;
				RTjpeg_yuv420rgb16(codec->decompress_struct, codec->dplanes, codec->row_pointers);
				break;
			case BC_RGB888:
				for(i = 0; i < h; i++)
					codec->row_pointers[i] = output + i * w * 3;
				RTjpeg_yuv420rgb24(codec->decompress_struct, codec->dplanes, codec->row_pointers);
				break;
			case BC_RGBA8888:
				for(i = 0; i < h; i++)
					codec->row_pointers[i] = output + i * w * 4;
				RTjpeg_yuv420rgb32(codec->decompress_struct, codec->dplanes, codec->row_pointers);
				break;
			case BC_BGR888:
				for(i = 0; i < h; i++)
					codec->row_pointers[i] = output + i * w * 3;
				RTjpeg_yuv420bgr24(codec->decompress_struct, codec->dplanes, codec->row_pointers);
				break;
			case BC_BGR8888:
				for(i = 0; i < h; i++)
					codec->row_pointers[i] = output + i * w * 4;
				RTjpeg_yuv420bgr32(codec->decompress_struct, codec->dplanes, codec->row_pointers);
				break;
		}
	}
	return result;
}

void quicktime_init_codec_rtjpeg(quicktime_video_map_t *vtrack)
{
	quicktime_rtjpeg_codec_t *codec;
	int i, w, h;

	w = vtrack->track->tkhd.track_width;
	h = vtrack->track->tkhd.track_height;

	((quicktime_codec_t*)vtrack->codec)->priv = calloc(1, sizeof(quicktime_rtjpeg_codec_t));
	codec = ((quicktime_codec_t*)vtrack->codec)->priv;
	((quicktime_codec_t*)vtrack->codec)->reads_colormodel = quicktime_reads_colormodel_rtjpeg;
	((quicktime_codec_t*)vtrack->codec)->writes_colormodel = quicktime_writes_colormodel_rtjpeg;

	codec->quality = 100;
	codec->compress_struct = RTjpeg_init();
	i = RTJ_YUV420;
	RTjpeg_set_format(codec->compress_struct, &i);
	RTjpeg_set_size(codec->compress_struct, &w, &h);
	i = (int)((float)codec->quality / 100.0 * 255.0);
	RTjpeg_set_quality(codec->compress_struct, &i);

	codec->decompress_struct = RTjpeg_init();
	i = RTJ_YUV420;
	RTjpeg_set_format(codec->decompress_struct, &i);

	codec->output_buffer = malloc(w*h*2);
	codec->input_buffer = malloc(w*h*2);
	codec->cplanes[0] = malloc(w*h+(w*h)/2);
	codec->cplanes[1] = codec->cplanes[0]+w*h; 
	codec->cplanes[2] = codec->cplanes[1]+(w*h)/4; 
	codec->dplanes[0] = malloc(w*h+(w*h)/2);
	codec->dplanes[1] = codec->dplanes[0]+w*h; 
	codec->dplanes[2] = codec->dplanes[1]+(w*h)/4; 
	codec->row_pointers = malloc(sizeof(unsigned char *) * h);
//	fprintf(stderr, "\nINIT %ld\n", (unsigned long int)codec);
}

void quicktime_codec_register(quicktime_extern_video_t * vcodec)
{
	strcpy(vcodec->fourcc, "RTJ0");
	vcodec->init = quicktime_init_codec_rtjpeg;
	vcodec->encode = qt_extern_encode;
	vcodec->decode = qt_extern_decode;
	vcodec->delete_codec = quicktime_delete_codec_rtjpeg;
}
